#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define RR_QUANTUM 2
#define CNTXT_SWITCH 1
#define MAX_PROCESSES 10 

enum pstate {
    WAITING,
    READY,
    FINISHED
};

struct pinfo {
    int id;
    int arrival_time;
    int execution_time;
    int priority;
    
    int wait_time;
    int turnaround_time;
    int remaining_time;
    int nb_time_pre_empted;  // Ajouté pour compter les préemptions
    enum pstate state;

    struct pinfo * next_pinfo;
};

struct perf_info {
    int total_time;
    int total_nr_ctxt_switch;
    int total_time_ctxt_switch;
};

void print_perf(struct perf_info * perf) {
    printf("Total time: %d\n", perf->total_time);
    printf("Total number of context switches: %d\n", perf->total_nr_ctxt_switch);
    printf("Total time spent on context switching: %d\n", perf->total_time_ctxt_switch);
}

void print_pinfo(struct pinfo * info) {
    printf("<Process {\n");
    printf("  PID: %d\n", info->id);
    printf("  Arrival time: %d\n", info->arrival_time);
    printf("  Execution time: %d\n", info->execution_time);
    printf("  Priority: %d\n", info->priority);
    printf("  Wait time: %d\n", info->wait_time);
    printf("  Turnaround time: %d\n", info->turnaround_time);
    printf("  Remaining time: %d\n", info->remaining_time);
    printf("  NEXT -> %p\n", info->next_pinfo);
    printf("}>\n");
}

void print_processes(struct pinfo * processes) {
    while (processes != NULL) {
        print_pinfo(processes);
        processes = processes->next_pinfo;
    }
}

struct pinfo * create_process(int id, int arrival_time, int execution_time, int priority) {
    struct pinfo * info = malloc(sizeof(struct pinfo));
    info->id = id;
    info->arrival_time = arrival_time;
    info->execution_time = execution_time;
    info->priority = priority;
    info->wait_time = 0;
    info->turnaround_time = 0;
    info->remaining_time = execution_time;
    info->nb_time_pre_empted = 0; // Initialiser le nombre de préemptions
    info->state = WAITING;
    info->next_pinfo = NULL;
    return info;
}

// Méthode de planification FCFS
struct perf_info schedule_FCFS(struct pinfo * processes) {
    int current_time = 0;
    struct pinfo * process = processes;

    struct perf_info perf = {0, 0, 0};
    
    while (process != NULL) {
        int wait_time = current_time - process->arrival_time;
        if (wait_time < 0) {
            wait_time = 0;
        }
        process->wait_time = wait_time;
        process->turnaround_time = process->execution_time + process->wait_time;
        current_time = process->arrival_time + process->turnaround_time;

        process = process->next_pinfo;
    }
    perf.total_time = current_time;
    
    return perf;
}

struct perf_info schedule_RR(struct pinfo *processes) { // Déclaration de la fonction schedule_RR qui prend un pointeur vers une liste de processus
    struct perf_info perf = {0, 0, 0}; // Initialisation de la structure de performance avec des valeurs à zéro
    int current_time = 0; // Variable pour suivre le temps actuel
    int finished_processes = 0; // Compteur pour le nombre de processus terminés

    while (finished_processes < MAX_PROCESSES) { // Boucle principale jusqu'à ce que tous les processus soient terminés
        int process_found = 0; // Indicateur pour savoir si un processus prêt a été trouvé
        struct pinfo *current_process = processes; // Pointeur pour parcourir la liste des processus

        while (current_process != NULL) { // Boucle pour parcourir tous les processus
            // Vérifiez si le processus est prêt à s'exécuter
            if (current_process->state != FINISHED && current_process->arrival_time <= current_time) { // Vérifie si le processus n'est pas fini et est arrivé
                process_found = 1; // Un processus prêt à s'exécuter a été trouvé

                int time_slice = (current_process->remaining_time < RR_QUANTUM) ? current_process->remaining_time : RR_QUANTUM; // Calcule la tranche de temps à exécuter

                // Simuler l'exécution
                current_time += time_slice; // Incrémente le temps actuel par la tranche de temps
                current_process->remaining_time -= time_slice; // Diminue le temps restant du processus

                // Calculer les temps d'attente pour les autres processus
                struct pinfo *other_process = processes; // Pointeur pour parcourir à nouveau la liste des processus
                while (other_process != NULL) { // Boucle pour parcourir tous les autres processus
                    if (other_process->state != FINISHED && other_process != current_process && other_process->arrival_time <= current_time) { // Vérifie si l'autre processus est prêt
                        other_process->wait_time += time_slice; // Augmente le temps d'attente des autres processus
                    }
                    other_process = other_process->next_pinfo; // Passe au processus suivant
                }

                // Gérer les statistiques de préemption
                if (current_process->remaining_time == 0) { // Vérifie si le processus est terminé
                    current_process->state = FINISHED; // Met à jour l'état du processus à fini
                    finished_processes++; // Incrémente le compteur de processus terminés
                    current_process->turnaround_time = current_time - current_process->arrival_time; // Calcule le temps de turnaround
                } else {
                    // Incrémenter le nombre de préemptions
                    current_process->nb_time_pre_empted++; // Incrémente le compteur de préemptions pour le processus actuel
                    perf.total_nr_ctxt_switch++; // Incrémente le nombre total de commutations de contexte
                }

                // Débogage : Afficher les informations du processus
                printf("Processus %d: remaining_time=%d, nb_time_pre_empted=%d\n", // Affiche les informations de débogage pour le processus actuel
                       current_process->id, current_process->remaining_time, current_process->nb_time_pre_empted);
            }

            current_process = current_process->next_pinfo; // Passe au processus suivant dans la liste
        }

        if (!process_found) { // Vérifie si aucun processus prêt n'a été trouvé
            // Aucun processus prêt, avancer le temps
            current_time++; // Incrémente le temps actuel si aucun processus n'est prêt
        }
    }

    perf.total_time = current_time; // Enregistre le temps total écoulé dans la structure de performance
    return perf; // Renvoie la structure de performance
}


void write_file(struct pinfo * process, struct perf_info * perf) {
    FILE *myStream_execution = fopen("executionRR.csv", "w");
    FILE *myStream_performance = fopen("performanceRR.csv", "w");

    if (myStream_execution == NULL || myStream_performance == NULL) {
        perror("Erreur à l'ouverture des fichiers");
        return;
    }

    while (process != NULL) {
        fprintf(myStream_execution, "%d,%d,%d,%d\n", 
                process->id,
                process->turnaround_time,
                process->wait_time,
                process->nb_time_pre_empted);
        process = process->next_pinfo;
    }
    fclose(myStream_execution);

    fprintf(myStream_performance, "%d,%d,%d\n",
            perf->total_time,
            perf->total_nr_ctxt_switch,
            perf->total_time_ctxt_switch);
    
    fclose(myStream_performance);
}







struct pinfo * read_file() {
    FILE * file = fopen("tasks.csv", "r");
    unsigned long buf_size = sizeof(char) * 64;
    char * line = (char *) malloc(buf_size);
    char * pid_str;
    char * arrival_str;
    char * execution_str;
    char * prio_str;

    struct pinfo * first = NULL;
    struct pinfo * last = NULL;
    struct pinfo * process;

    while (fgets(line, buf_size, file)) {
        pid_str = strtok(line, " ");
        arrival_str = strtok(NULL, " ");
        execution_str = strtok(NULL, " ");
        prio_str = strtok(NULL, " ");

        process = create_process(
            atoi(pid_str),
            atoi(arrival_str),
            atoi(execution_str),
            atoi(prio_str)
        );

        // Si la liste n'est pas initialisée
        if (first == NULL) {
            first = process;
        }

        // Si un élément est déjà dans la liste
        if (last != NULL) {
            last->next_pinfo = process;
        }
        last = process;

        print_pinfo(process);
    }
    free(line);
    fclose(file);
    return first;
}

void free_processes(struct pinfo * next) {
    struct pinfo * cur;
    while (next != NULL) {
        cur = next;
        next = cur->next_pinfo;
        free(cur);
    }
}



int main() {
    struct pinfo * processes = read_file();

    //struct perf_info perf = schedule_FCFS(processes);  // Remise en place de FCFS
    struct perf_info perf = schedule_RR(processes); 

    write_file(processes, &perf);

    free_processes(processes);

    return 0;
}
